@@ -6,7 +6,7 @@
  * This header file is created from pg_stat_statements.c to implement
  * normalization of query string.
  *
- * Portions Copyright (c) 2008-2017, PostgreSQL Global Development Group
+ * Portions Copyright (c) 2008-2014, PostgreSQL Global Development Group
  */
 #ifndef NORMALIZE_QUERY_H
 #define NORMALIZE_QUERY_H
@@ -40,14 +40,11 @@ typedef struct pgssJumbleState
 
 	/* Current number of valid entries in clocations array */
 	int			clocations_count;
-
-	/* highest Param id we've seen, in order to start normalization correctly */
-	int			highest_extern_param_id;
 } pgssJumbleState;
 
 static char *
 generate_normalized_query(pgssJumbleState *jstate, const char *query,
-						  int query_loc, int *query_len_p, int encoding);
+						  int *query_len_p, int encoding);
 static void JumbleQuery(pgssJumbleState *jstate, Query *query);
 
 #define JUMBLE_SIZE		1024
